/* fueue.c -- FIFO XORed IPC File Queue
 * 
 * Copyright (C) 2004 H Zero Seven
 *
 * Authors: Christian Volmering <risk@h07.org>
 * 
 * This file is part of the aexpl-ng distribution.
 * 
 * The aexpl-ng distribution is free software; you can redistribute it 
 * and/or modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 of 
 * the License, or (at your option) any later version.
 *
 * The aexpl-ng distribution is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the aexpl-ng distribution; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 * MA 02111-1307 USA
 */

#include "config.h"

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>

#ifdef WITH_DMALLOC
#  include <dmalloc.h>
#endif

#if defined _SEM_SEMUN_UNDEFINED
union semun
{
  int val;
  struct semid_ds *buf;
  unsigned short *array;
  struct seminfo *__buf;
};
#endif

struct _node_t
{
  char file[PATH_MAX];
  unsigned long link;
};

struct _queue_t
{
  unsigned int size;
  unsigned int max_size;
  unsigned int __pos;
  off_t __off_base;
  off_t __off_tail;
  off_t __off_next;
};

struct _fqueue_t
{
  int shm_id;
  int sem_id;
  void *__buf;
};

#define __SEM_MEM   0
#define __SEM_QUEUE 1

static struct sembuf __sem_mem_lock = { __SEM_MEM, -1, SEM_UNDO };
static struct sembuf __sem_mem_unlock = { __SEM_MEM, +1, SEM_UNDO };
static struct sembuf __sem_wait_full = { __SEM_QUEUE, -1, SEM_UNDO };
static struct sembuf __sem_set_full = { __SEM_QUEUE, +1, SEM_UNDO };

static int _sem_operate (struct _fqueue_t *fqueue, struct sembuf *buf);
static struct _node_t *_new_node (struct _fqueue_t *fqueue);
static inline struct _queue_t *_get_queue (struct _fqueue_t *fqueue);
static inline void *_get_ptr (struct _fqueue_t *fqueue, off_t offset);
static inline off_t _get_offset (struct _fqueue_t *fqueue, void *ptr);
static inline struct _node_t *_get_node (struct _fqueue_t *fqueue,
					 off_t offset);
static int _fempty (struct _fqueue_t *fqueue);
static int _fsize (struct _fqueue_t *fqueue);
static void _fenqueue (struct _fqueue_t *fqueue, const char *file);
static char *_fdequeue (struct _fqueue_t *fqueue);

/* {{{ _sem_operate */
static int
_sem_operate (struct _fqueue_t *fqueue, struct sembuf *buf)
{
  int result;

  while (((result = semop (fqueue->sem_id, buf, 1)) == -1)
	 && (errno == EINTR));

  return (result ? 0 : 1);
}

/* }}} */

/* {{{ _get_ptr */
static inline void *
_get_ptr (struct _fqueue_t *fqueue, off_t offset)
{
  return ((void *) (fqueue->__buf + offset));
}

/* }}} */

/* {{{ _get_offset */
static inline off_t
_get_offset (struct _fqueue_t *fqueue, void *ptr)
{
  return ((off_t) (ptr - fqueue->__buf));
}

/* }}} */

/* {{{ _get_queue */
static inline struct _queue_t *
_get_queue (struct _fqueue_t *fqueue)
{
  return ((struct _queue_t *) fqueue->__buf);
}

/* }}} */

/* {{{ _get_node */
static inline struct _node_t *
_get_node (struct _fqueue_t *fqueue, off_t offset)
{
  return ((struct _node_t *) _get_ptr (fqueue, offset));
}

/* }}} */

/* {{{ _new_node */
static struct _node_t *
_new_node (struct _fqueue_t *fqueue)
{
  struct _queue_t *queue;
  struct _node_t *node;

  queue = _get_queue (fqueue);

  if (queue->max_size == queue->__pos)
    {
      queue->__off_next = (off_t) sizeof (struct _queue_t);
      queue->__pos = 0;
    }
  node = _get_node (fqueue, queue->__off_next);

  queue->__off_next += (off_t) sizeof (struct _node_t);
  queue->__pos++;

  return (node);
}

/* }}} */

/* {{{ _fempty */
static int
_fempty (struct _fqueue_t *fqueue)
{
  return (_get_queue (fqueue)->__off_base ? 0 : 1);
}

/* }}} */

/* {{{ _fsize */
static int
_fsize (struct _fqueue_t *fqueue)
{
  return (_get_queue (fqueue)->size);
}

/* }}} */

/* {{{ _fenqueue */
void
_fenqueue (struct _fqueue_t *fqueue, const char *file)
{
  struct _queue_t *queue;
  struct _node_t *node;
  off_t off_node;

  queue = _get_queue (fqueue);
  node = _new_node (fqueue);
  off_node = _get_offset (fqueue, (void *) node);

  if (queue->max_size == queue->size)
    _fdequeue (fqueue);

  snprintf (node->file, PATH_MAX, file);
  node->link = 0;

  if (!_fempty (fqueue))
    {
      node->link ^= (long) queue->__off_tail;
      _get_node (fqueue, queue->__off_tail)->link ^= (long) off_node;
      queue->__off_tail = off_node;
    }
  else
    queue->__off_tail = queue->__off_base = off_node;

  queue->size++;
}

/* }}} */

/* {{{ _fdequeue */
static char *
_fdequeue (struct _fqueue_t *fqueue)
{
  struct _queue_t *queue = _get_queue (fqueue);
  struct _node_t *node = NULL;

  if (queue->__off_base)
    {
      node = _get_node (fqueue, queue->__off_base);

      if ((queue->__off_base = node->link ^ 0))
	_get_node (fqueue, queue->__off_base)->link ^=
	  (long) _get_offset (fqueue, (void *) node);
      else
	queue->__off_tail = 0;

      queue->size--;
    }

  return (node ? node->file : NULL);
}

/* }}} */

/* {{{ new_fqueue */
struct _fqueue_t *
new_fqueue (int max_size)
{
  struct _fqueue_t *fqueue;
  struct _queue_t *queue;
  union semun un;
  size_t shm_size = sizeof (struct _queue_t) +
    sizeof (struct _node_t) * max_size;

  void _cleanup (void)
  {
    if (fqueue->sem_id != -1)
      {
	semctl (fqueue->sem_id, __SEM_MEM, IPC_RMID, NULL);
	semctl (fqueue->sem_id, __SEM_QUEUE, IPC_RMID, NULL);
      }

    if (fqueue->shm_id != -1)
      {
	shmctl (fqueue->shm_id, IPC_RMID, NULL);
      }

    free (fqueue);
  }

  if (!(fqueue = (struct _fqueue_t *) malloc (sizeof (struct _fqueue_t))))
    return (NULL);

  fqueue->sem_id = -1;
  fqueue->shm_id = -1;

  if ((fqueue->sem_id = semget (IPC_PRIVATE, 2, 0600 | IPC_CREAT)) == -1)
    {
      _cleanup ();

      return (NULL);
    }

  un.val = 1;
  if (semctl (fqueue->sem_id, __SEM_MEM, SETVAL, un) == -1)
    {
      _cleanup ();

      return (NULL);
    }

  un.val = 0;
  if (semctl (fqueue->sem_id, __SEM_QUEUE, SETVAL, un) == -1)
    {
      _cleanup ();

      return (NULL);
    }

  if ((fqueue->shm_id =
       shmget (IPC_PRIVATE, shm_size, 0600 | IPC_CREAT)) == -1)
    {
      _cleanup ();

      return (NULL);
    }

  if ((fqueue->__buf = shmat (fqueue->shm_id, 0, 0)) == (void *) -1)
    {
      _cleanup ();

      return (NULL);
    }

  queue = _get_queue (fqueue);

  queue->size = 0;
  queue->max_size = max_size;
  queue->__pos = 0;
  queue->__off_base = queue->__off_tail = 0;
  queue->__off_next = (off_t) sizeof (struct _queue_t);

  return (fqueue);
}

/* }}} */

/* {{{ open_fqueue */
int
open_fqueue (struct _fqueue_t *fqueue)
{
  fqueue->__buf = shmat (fqueue->shm_id, 0, 0);
  if (fqueue->__buf == (void *) -1)
    return (-1);

  return (1);
}

/* }}} */

/* {{{ free_fqueue */
void
free_fqueue (struct _fqueue_t *fqueue)
{
  if (!_sem_operate (fqueue, &__sem_mem_lock))
    return;

  shmctl (fqueue->shm_id, IPC_RMID, NULL);
  semctl (fqueue->sem_id, __SEM_MEM, IPC_RMID, NULL);
  semctl (fqueue->sem_id, __SEM_QUEUE, IPC_RMID, NULL);

  free (fqueue);
}

/* }}} */

/* {{{ fempty */
int
fempty (struct _fqueue_t *fqueue)
{
  unsigned short empty;

  if (!_sem_operate (fqueue, &__sem_mem_lock))
    return (-1);

  empty = _fempty (fqueue);

  _sem_operate (fqueue, &__sem_mem_unlock);

  return (empty);
}

/* }}} */

/* {{{ fsize */
int
fsize (struct _fqueue_t *fqueue)
{
  unsigned int size;

  if (!_sem_operate (fqueue, &__sem_mem_lock))
    return (-1);

  size = _fsize (fqueue);

  _sem_operate (fqueue, &__sem_mem_unlock);

  return (size);
}

/* }}} */

/* {{{ fclear */
int
fclear (struct _fqueue_t *fqueue)
{
  int i;

  if (!_sem_operate (fqueue, &__sem_mem_lock))
    return (-1);

  for (i = _fsize (fqueue); i > 0; --i)
    _fdequeue (fqueue);

  _sem_operate (fqueue, &__sem_mem_unlock);

  return (1);
}

/* }}} */

/* {{{ fenqueue */
int
fenqueue (struct _fqueue_t *fqueue, const char *file)
{
  unsigned short empty;

  if (!_sem_operate (fqueue, &__sem_mem_lock))
    return (-1);

  empty = _fempty (fqueue);
  _fenqueue (fqueue, file);

  if (empty)
    _sem_operate (fqueue, &__sem_set_full);

  _sem_operate (fqueue, &__sem_mem_unlock);

  return (1);
}

/* }}} */

/* {{{ fdequeue */
char *
fdequeue (struct _fqueue_t *fqueue, char *buf)
{
  char *de_file, *file = NULL;

  if (!_sem_operate (fqueue, &__sem_wait_full))
    return (NULL);

  if (!_sem_operate (fqueue, &__sem_mem_lock))
    return (NULL);

  if (!(de_file = _fdequeue (fqueue)))
    return (NULL);

  if (!buf)
    file = (char *) calloc (strlen (de_file), sizeof (char));
  else
    file = buf;

  if (!file)
    return (NULL);

  strcpy (file, de_file);

  if (!_fempty (fqueue))
    _sem_operate (fqueue, &__sem_set_full);

  _sem_operate (fqueue, &__sem_mem_unlock);

  return (file);
}

/* }}} */

/*
 * Local variables:
 * tab-width: 8
 * c-basic-offset: 8
 * End:
 * vim600: noet nonu sw=8 ts=8 fdm=marker
 * vim<600: noet nonu sw=8 ts=8
 */
